"""Main entrypoint."""

import logging
import os
from typing import Callable

from typer import Option, Typer

from llmask.model import get_api_client
from llmask.transform import (
    chain_apply_transformations,
    parse_transformations_string,
)

logging.basicConfig(level=logging.WARN)
os.environ["NO_COLOR"] = "1"  # deactivate color for rich/colorama (if installed)

app = Typer(
    name="llmask",
    add_completion=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@app.command()
def transform(
    transformations: str = Option(
        "ts",
        "-t",
        "--transformations",
        help=(
            "Compact representation of operations "
            "(e.g. 'ts' for the steps 'thesaurus -> simplify')"
        ),
    ),
    input: str = Option(
        ...,
        "-i",
        "--input",
        help="Input text that will be transformed.",
    ),
    model_name: str = Option(
        "nous-hermes2:10.7b-solar-q6_K",
        "-m",
        "--model",
        help="Name of model to use (as known to model server).",
    ),
    url: str = Option(
        "http://localhost:11434/v1",
        "-u",
        "--url",
        help="URL of Open AI compatible model API.",
    ),
):
    """Transform input text with chained transformations by a Large Language Model."""
    print("\nUser-provided input:\n")
    print(f"> {input}\n\n")

    transformation_funcs: list[Callable] = parse_transformations_string(
        transformations=transformations
    )

    api_client = get_api_client(url=url)
    transformed_texts = chain_apply_transformations(
        input=input,
        transformation_funcs=transformation_funcs,
        model_name=model_name,
        api_client=api_client,
    )

    for func, text in zip(transformation_funcs, transformed_texts):
        print(f"Result after applying transformation '{func.__name__}':\n")
        print(f"> {text}\n\n")


if __name__ == "__main__":
    app()



"""Module with functions to transform text."""

from typing import Callable

from openai import OpenAI

from llmask.model import query_llm


# OPTIONAL: make functions retun system prompts only, and move queries LLM in model.py?
def thesaurus(
    input: str,
    model_name: str,
    api_client: OpenAI,
) -> str:
    """Change input by replacing words with their synonyms.

    Args:
        input: Text input to be transformed.
        model_name: name of model to use (as known to model server).
        api_client: instance of adapter to model API.
    Return:
        Transformed text.
    """
    instructions: str = """
        change the user input by replacing each word with their synonyms.
        replace every word. change all words.
        find a word with similar meaning and use it in the original word's place.

        think like a thesaurus, replacing each word.
        but do not change the meaning of the paragraph.
        Only output the changed input.
        Do not start with 'understood' etc.
    """
    response = query_llm(
        api_client=api_client,
        instructions=instructions,
        input=input,
        model_name=model_name,
        temperature=1.5,
        seed=42,
    )
    return response


def simplify(
    input: str,
    model_name: str,
    api_client: OpenAI,
) -> str:
    """Simplify language of input.

    Args:
        input: Text input to be transformed.
        model_name: name of model to use (as known to model server).
        api_client: instance of adapter to model API.
    Return:
        Transformed text.
    """
    instructions: str = """
        change the user input by simplifying the language.
        replace uncommon words by common ones.
        make the writing as simple as possible.
        make sure that even children would know each word.

        Do not change the input's meaning.
        do not leave out aspects.
        it's important that the output means the same as the input.

        Only output the changed input.
        Do not start with 'understood' etc.
    """
    response = query_llm(
        api_client=api_client,
        instructions=instructions,
        model_name=model_name,
        input=input,
        temperature=0.3,
        seed=42,
    )

    return response


TRANSFORMATION_MAPPING = {
    "s": simplify,
    "t": thesaurus,
}


def parse_transformations_string(transformations: str) -> list[Callable]:
    """Parse CLI parameter 'transformations' to list of to-be-applied functions.

    Args:
        transformstions: sequential transformations, in compact string format.
    """
    # check input validity
    for transformation in transformations:
        try:
            assert transformation in TRANSFORMATION_MAPPING.keys()
        except AssertionError:
            print(
                f"Transformation key '{transformation}' is not valid! "
                f"Valid options are: {list(TRANSFORMATION_MAPPING.keys())}. "
                "Exiting."
            )
            exit(1)

    return [TRANSFORMATION_MAPPING[char] for char in transformations]


def chain_apply_transformations(
    input: str,
    transformation_funcs: list[Callable],
    model_name: str,
    api_client: OpenAI,
) -> list[str]:
    """Apply chain of transformations, passing each result as input to the next step.

    Args:
        input: user-provided text input.
        transformation_funcs: list of transformation functions to be chained.
        model_name: name of model to use (as known to model server).
        api_client: instance of adapter to model API.
    Return:
        list of transformed text, for each step of transformation pipeline.
    """
    transformed_texts: list[str] = []  # list with each result of transformation chain
    for transformation_func in transformation_funcs:
        output = transformation_func(
            input=input,
            model_name=model_name,
            api_client=api_client,
        )
        transformed_texts.append(output)
        input = output  # use this transformtion's output as next input
    return transformed_texts



"""Wrapper for prompting LLM."""

# %%
from openai import OpenAI


def get_api_client(url: str) -> OpenAI:
    """Get API client to locally running LLM, which has OpenAI compatibility.

    Args:
        url: URL to OpenAI compatible model API.
    Return:
        Instance of client for model API.
    """
    return OpenAI(
        base_url=url,
        api_key="sk-no-key-required",
    )


# OPTIONAL: pass down model name
def query_llm(
    api_client: OpenAI,
    instructions: str,
    input: str,
    model_name: str,
    temperature: float,
    seed: int,
) -> str:
    """Interface function to query LLM and retrieve response.

    Args:
        api_client: Instance of client for model API
        instructions: instructions on how to change input text
        input: input text to be changed
        model_name: name of LLM to be used (known to model server)
        temperature: parameter passed to LLM
        seed: random seed (for reproducibility)
    Returns:
        Response text from LLM
    """
    completion = api_client.chat.completions.create(
        model=model_name,
        messages=[
            {
                "role": "system",
                "content": instructions,
            },
            {
                "role": "user",
                "content": input,
            },
        ],
        temperature=temperature,
        seed=seed,
    )

    response = str(completion.choices[0].message.content)
    return response


# %%
